#!/usr/bin/env python3
import os
import sys
import shutil
import numpy as np
import lx.tools
import lx.parser


##GENERATES NEUTRAL INPUT######################################
def gera_optcom(atomos, G, base, nproc, mem, omega, op):
    header = f"%nproc=JJJ\n%mem=MEM\n# BASE iop(3/108=MMMMM00000) iop(3/107=MMMMM00000) {op}\n\nTITLE\n\n0 1\n"
    header = (
        header.replace("JJJ", nproc)
        .replace("MEM", mem)
        .replace("BASE", base)
        .replace("MMMMM", omega)
    )
    file = "OPT_" + omega + "_.com"
    lx.tools.write_input(atomos, G, header, "", file)
    return file


###############################################################


##GENERATES ION INPUTS#########################################
def gera_ioncom(atomos, G, base, nproc, mem, omega):
    header = "%nproc=JJJ\n%mem=MEM\n# BASE iop(3/108=MMMMM00000) iop(3/107=MMMMM00000)\n\nTITLE\n\n1 2\n"
    header = (
        header.replace("JJJ", nproc)
        .replace("MEM", mem)
        .replace("BASE", base)
        .replace("MMMMM", omega)
    )
    file1 = "pos_" + omega + "_.com"
    file2 = "neg_" + omega + "_.com"
    lx.tools.write_input(atomos, G, header, "", file1)
    header = "%nproc=JJJ\n%mem=MEM\n# BASE iop(3/108=MMMMM00000) iop(3/107=MMMMM00000) \n\nTITLE\n\n-1 2\n"
    header = (
        header.replace("JJJ", nproc)
        .replace("MEM", mem)
        .replace("BASE", base)
        .replace("MMMMM", omega)
    )
    lx.tools.write_input(atomos, G, header, "", file2)
    return [file1, file2]


###############################################################


##GETS ENERGY FROM LOG#########################################
def pega_energia(file):
    energias = []
    with open(file, "r",encoding='utf-8') as f:
        for line in f:
            if "SCF Done:" in line:
                line = line.split()
                energias.append(float(line[4]))
    return min(energias)


###############################################################


##MAP TO FLOAT WITH EXCEPTION##################################
def to_float(num):
    try:
        num = float(num)
    except ValueError:
        num = -np.inf
    return num


###############################################################


##GETS HOMO ENERGY FROM LOG####################################
def pega_homo(file):
    HOMOS = []
    with open(file, "r",encoding='utf-8') as f:
        for line in f:
            if "OPT" in file and "Optimized Parameters" in line:
                HOMOS = []
            if "occ. eigenvalues" in line:
                line = line.split()
                homos = line[4:]
                HOMOS.extend(homos)
    if len(HOMOS) == 0:
        with open(file, "r",encoding='utf-8') as f:
            for line in f:
                if "occ. eigenvalues" in line:
                    line = line.split()
                    homos = line[4:]
                    HOMOS.extend(homos)
    HOMOS = list(map(to_float, HOMOS))
    return np.max(HOMOS)


###############################################################


##RUNS CALCULATIONS############################################
def rodar_omega(atomos, geom, base, nproc, mem, omega, op, batch_file, gaussian, numjobs):
    omega = f"{omega:05.0f}"
    file = gera_optcom(atomos, geom, base, nproc, mem, omega, op)
    remover = []
    if op == "opt":
        the_watcher = lx.tools.Watcher('.',files=[file])
        the_watcher.run(batch_file, gaussian, 1)
        the_watcher.hold_watch()
        geom, atomos = lx.parser.pega_geom(file[:-3] + "log")
        files = gera_ioncom(atomos, geom, base, nproc, mem, omega)
        the_watcher = lx.tools.Watcher('.',files=files)
        the_watcher.run(batch_file, gaussian, min(numjobs,2))
        the_watcher.hold_watch()
    else:
        files = gera_ioncom(atomos, geom, base, nproc, mem, omega)
        the_watcher = lx.tools.Watcher('.',files=[file]+files)
        the_watcher.run(batch_file, gaussian, min(numjobs,3))
        the_watcher.hold_watch()

    logs = files + [file]
    for file in logs:
        remover.append(file)
        if "pos_" in file:
            cation = pega_energia(file[:-3] + "log")
        elif "neg_" in file:
            anion = pega_energia(file[:-3] + "log")
            homo_anion = pega_homo(file[:-3] + "log")
        elif "OPT_" in file:
            neutro = pega_energia(file[:-3] + "log")
            homo_neutro = pega_homo(file[:-3] + "log")

    try:
        os.mkdir("Logs")
    except FileExistsError:
        pass
    for file in remover:
        shutil.move(file, "Logs/" + file)
        shutil.move(file[:-3] + "log", "Logs/" + file[:-3] + "log")
    J = np.sqrt(
        ((homo_neutro + cation - neutro) ** 2 + (homo_anion + neutro - anion) ** 2)
    ) * (27.2114)
    return J, geom, atomos


###############################################################


##WRITES LOG WITH RESULTS######################################
def write_tolog(omegas, Js, frase):
    with open("omega.lx", "w",encoding='utf-8') as f:
        f.write("#w(10^4 bohr^-1)    J(eV)\n")
        list1, list2 = zip(*sorted(zip(omegas, Js)))
        for i in range(len(list1)):
            f.write(f"{list1[i]:05.0f}               {list2[i]:.4f}\n")
        f.write(f"\n{frase} {list1[list2.index(min(list2))]:05.0f}\n")


###############################################################


def main():
    geomlog = sys.argv[1]
    base = sys.argv[2]
    nproc = sys.argv[3]
    mem = sys.argv[4]
    omega1 = sys.argv[5]
    passo = sys.argv[6]
    relax = sys.argv[7]
    script = sys.argv[8]
    gaussian = sys.argv[9]
    parallel = sys.argv[10]
    try:
        int(nproc)
        passo = float(passo) * 10000
        omega1 = float(omega1) * 10000
    except ValueError:
        lx.parser.fatal_error("nproc, omega and step must be numbers. Goodbye!")
    if relax.lower() == "y":
        op = "opt"
    elif relax.lower() == "n":
        op = ""
    else:
        lx.parser.fatal_error("It must be either y or n. Goodbye!")
    if parallel.lower() == "y":
        numjobs = 10
    else:
        numjobs = 1
    omegas, Js = [], []
    oms, jotas = [], []
    try:
        with open("omega.lx", "r",encoding='utf-8') as f:
            for line in f:
                line = line.split()
                if len(line) == 2 and "#" not in line:
                    om = float(line[0])
                    omegas.append(om)
                    Js.append(float(line[1]))
        menor = omegas[Js.index(min(Js))]
        G, atomos = lx.parser.pega_geom(f"Logs/OPT_{menor:05.0f}_.log")
    except FileNotFoundError:
        G, atomos = lx.parser.pega_geom(geomlog)

    while passo > 25:
        if omega1 in omegas:
            ind = omegas.index(omega1)
            J = Js[ind]
        else:
            J, G, atomos = rodar_omega(
                atomos, G, base, nproc, mem, omega1, op, script, gaussian, numjobs
            )
            omegas.append(omega1)
            Js.append(J)
        oms.append(omega1)
        jotas.append(J)
        try:
            if jotas[-1] - jotas[-2] > 0:
                passo = int(passo / 2)
                sign = -1 * np.sign(int(oms[-1]) - int(oms[-2]))
            else:
                sign = +1 * np.sign(int(oms[-1]) - int(oms[-2]))
            omega1 += sign * passo

        except IndexError:
            omega1 += passo
        write_tolog(omegas, Js, "#Best value so far:")

    write_tolog(omegas, Js, "#Done! Optimized value:")
    menor = omegas[Js.index(min(Js))]
    log = f"Logs/OPT_{menor:05.0f}_.log"
    G, atomos = lx.parser.pega_geom(log)
    base, _, nproc, mem, scrf, _ = lx.parser.busca_input(log)
    cm = lx.parser.get_cm(log)
    header = f"%nproc={nproc}\n%mem={mem}\n# {base} {scrf}\n\nTITLE\n\n{cm}\n"
    lx.tools.write_input(atomos, G, header, "", "tuned_w.com")


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
import os
import sys
import time
import requests
import pkg_resources
import subprocess
from scipy.stats import norm
import numpy as np
import pandas as pd
from lx.ld import run_ld
import lx.parser

##SOME CONSTANTS##############################################
EPSILON_0 = lx.parser.EPSILON_0  # F/m
HBAR_EV = lx.parser.HBAR_EV  # eV s
HBAR_J = lx.parser.HBAR_J  # J s
MASS_E = lx.parser.MASS_E  # kg
LIGHT_SPEED = lx.parser.LIGHT_SPEED  # m/s
E_CHARGE = lx.parser.E_CHARGE  # C
BOLTZ_EV = lx.parser.BOLTZ_EV  # eV/K
AMU = lx.parser.AMU  # kg
###############################################################


def distance_matrix(geom):
    matrix = np.zeros((1, np.shape(geom)[0]))
    for ind in range(np.shape(geom)[0]):
        distances = geom - geom[ind, :]
        distances = np.sqrt(np.sum(np.square(distances), axis=1))
        matrix = np.vstack((matrix, distances[np.newaxis, :]))
    matrix = matrix[1:, :]
    return matrix


def adjacency(geom, atoms):
    covalent_radii = {
        '1': 0.31,
        'H': 0.31,
        '2': 0.28,
        'He': 0.28,
        '3': 1.28,
        'Li': 1.28,
        '4': 0.96,
        'Be': 0.96,
        '5': 0.84,
        'B': 0.84,
        '6': 0.76,
        'C': 0.76,
        '7': 0.71,
        'N': 0.71,
        '8': 0.66,
        'O': 0.66,
        '9': 0.57,
        'F': 0.57,
        '10': 0.58,
        'Ne': 0.58,
        '11': 1.66,
        'Na': 1.66,
        '12': 1.41,
        'Mg': 1.41,
        '13': 1.21,
        'Al': 1.21,
        '14': 1.11,
        'Si': 1.11,
        '15': 1.07,
        'P': 1.07,
        '16': 1.05,
        'S': 1.05,
        '17': 1.02,
        'Cl': 1.02,
        '18': 1.06,
        'Ar': 1.06,
        '19': 2.03,
        'K': 2.03,
        '20': 1.76,
        'Ca': 1.76,
        '21': 1.7,
        'Sc': 1.7,
        '22': 1.6,
        'Ti': 1.6,
        '23': 1.53,
        'V': 1.53,
        '24': 1.39,
        'Cr': 1.39,
        '25': 1.61,
        'Mn': 1.61,
        '26': 1.52,
        'Fe': 1.52,
        '27': 1.50,
        'Co': 1.50,
        '28': 1.24,
        'Ni': 1.24,
        '29': 1.32,
        'Cu': 1.32,
        '30': 1.22,
        'Zn': 1.22,
        '31': 1.22,
        'Ga': 1.22,
        '32': 1.2,
        'Ge': 1.2,
        '33': 1.19,
        'As': 1.19,
        '34': 1.20,
        'Se': 1.20,
        '35': 1.20,
        'Br': 1.20,
        '36': 1.16,
        'Kr': 1.16,
        '37': 2.2,
        'Rb': 2.2,
        '38': 1.95,
        'Sr': 1.95,
        '39': 1.9,
        'Y': 1.9,
        '40': 1.75,
        'Zr': 1.75,
        '41': 1.64,
        'Nb': 1.64,
        '42': 1.54,
        'Mo': 1.54,
        '43': 1.47,
        'Tc': 1.47,
        '44': 1.46,
        'Ru': 1.46,
        '45': 1.42,
        'Rh': 1.42,
        '46': 1.39,
        'Pd': 1.39,
        '47': 1.45,
        'Ag': 1.45,
        '48': 1.44,
        'Cd': 1.44,
        '49': 1.42,
        'In': 1.42,
        '50': 1.39,
        'Sn': 1.39,
        '51': 1.39,
        'Sb': 1.39,
        '52': 1.38,
        'Te': 1.38,
        '53': 1.39,
        'I': 1.39,
        '54': 1.4,
        'Xe': 1.4,
    }
    dist_matrix = distance_matrix(geom)
    adj_matrix = np.zeros(np.shape(dist_matrix))
    # connectivity matrix
    for i in range(np.shape(dist_matrix)[0]):
        for j in range(i, np.shape(dist_matrix)[1]):
            r_e = (covalent_radii[atoms[i]] + covalent_radii[atoms[j]]) + 0.4
            if 0.8 < dist_matrix[i, j] < r_e:
                adj_matrix[i, j] = adj_matrix[j, i] = 1
    return adj_matrix


def fingerprint(file, folder):
    try:
        geom, atoms = lx.parser.pega_geom(folder + "/" + file)
        cm = adjacency(geom, atoms)
    except UnboundLocalError:
        cm = np.zeros((1, 1))
    return cm


##SAVES OPT GEOMETRY###########################################
def salva_geom(geom, atomos):
    atomos = np.array([atomos]).astype(float)
    atomos = atomos.T
    geom = np.hstack((atomos, geom))
    np.savetxt(
        "opt_geom.lx", geom, delimiter="\t", fmt=["%1.1u", "%+1.5f", "%+1.5f", "%+1.5f"]
    )
    print("The optimized geometry that is used is saved in the opt_geom.lx file!")


###############################################################


##WRITES ATOMS AND XYZ COORDS TO FILE##########################
def write_input(atomos, geom, header, bottom, file):
    with open(file, "w", encoding="utf-8") as f:
        f.write(header)
        for i, atomo in enumerate(atomos):
            texto = f"{atomo:2s}  {geom[i,0]:.6f}  {geom[i,1]:.6f}  {geom[i,2]:.6f}\n"
            f.write(texto)
        f.write("\n" + bottom + "\n")


###############################################################


##CHECKS FOR EXISTING GEOMETRIES###############################
def start_counter():
    files = [
        file
        for file in os.listdir("Geometries")
        if ".com" in file and "Geometr" in file
    ]
    return len(files)


###############################################################


##DISTORTS GEOMETRIES IN DIRECTION OF IMAG FREQ################
def distort(freqlog):
    temp = 300
    geom, atomos = lx.parser.pega_geom(freqlog)
    freqs, masses = lx.parser.pega_freq(freqlog)
    normal_coords = lx.parser.pega_modos(geom, freqlog)
    final_geom = geom.copy()
    # check for imaginary frequencies
    if np.all(freqs > 0):
        lx.parser.fatal_error("No imaginary frequencies found. Exiting...")
    for i, freq in enumerate(freqs):
        if freq < 0:
            f = -1 * freq
            q = 3 * np.sqrt(
                HBAR_J / (2 * masses[i] * f * np.tanh(HBAR_EV * f / (2 * BOLTZ_EV * temp)))
            )
            q = np.array(q)
            final_geom += q * normal_coords[:, :, i]
    base, _, nproc, mem, _, _ = lx.parser.busca_input(freqlog)
    cm = lx.parser.get_cm(freqlog)
    header = f"%nproc={nproc}\n%mem={mem}\n# {base} opt freq=noraman\n\nDISTORTED GEOM\n\n{cm}\n"
    bottom = "\n"
    write_input(atomos, final_geom, header, bottom, "distorted.com")
    print("New input file written to distorted.com")


###############################################################
def sample_geometries(freqlog, num_geoms, temp, limit=np.inf, warning=True, show_progress=False):
    geom, atomos = lx.parser.pega_geom(freqlog)
    old = adjacency(geom, atomos)
    freqs, masses = lx.parser.pega_freq(freqlog)
    normal_coord = lx.parser.pega_modos(geom, freqlog)
    # check for negative frequencies
    rejected_geoms = 0
    if warning:
        lx.parser.double_check(freqlog)
    else:
        freqs[freqs < 0] *= -1
        mask = freqs < limit * (LIGHT_SPEED * 100 * 2 * np.pi)
        freqs = freqs[mask]
        masses = masses[mask]
        normal_coord = normal_coord[:,:, mask]
    structures = np.zeros((geom.shape[0], geom.shape[1], num_geoms))
    scales = 1e10 * np.sqrt(
        HBAR_J / (2 * masses * freqs * np.tanh(HBAR_EV * freqs / (2 * BOLTZ_EV * temp)))
    )
    for j in range(num_geoms):
        ok = False
        while not ok:
            start_geom = geom.copy()
            qs = [norm(scale=scale, loc=0).rvs(size=1) for scale in scales]
            qs = np.array(qs)
            start_geom += np.sum(qs.reshape((1, 1, -1)) * normal_coord, axis=2)
            new = adjacency(start_geom, atomos)
            if 0.5 * np.sum(np.abs(old - new)) < 1 or not warning:
                ok = True
                structures[:, :, j] = start_geom
            else:
                rejected_geoms += 1
            if show_progress:
                progress = 100 * (j + 1) / num_geoms
                text = f"{progress:2.1f}%"
                print(" ", text, "of the geometries done.", rejected_geoms, "geometries rejected", end="\r", flush=True)
        try:
            numbers = np.vstack((numbers, qs.T))
        except UnboundLocalError:
            numbers = qs.T
    numbers = np.round(numbers, 4)
    return numbers, atomos, structures


###############################################################


##MAKES ENSEMBLE###############################################
def make_ensemble(freqlog, num_geoms, temp, header, bottom):
    try:
        os.mkdir("Geometries")
    except FileExistsError:
        pass
    counter = start_counter()
    print("\nGenerating geometries...\n")
    numbers, atomos, structures = sample_geometries(freqlog, num_geoms, temp, show_progress=True)
    freqs, masses = lx.parser.pega_freq(freqlog)
    # convert numbers to dataframe
    numbers = pd.DataFrame(
        numbers, columns=[f"mode_{i+1}" for i in range(np.shape(numbers)[1])]
    )
    # check if file exists
    if os.path.isfile(f"Magnitudes_{temp:.0f}K_.lx"):
        data = pd.read_csv(f"Magnitudes_{temp:.0f}K_.lx")
        # get only columns with mode_ in the name
        data = data.filter(regex="mode_")
        # remove nan values
        data = data.dropna()
        # join data and numbers on axis 0
        numbers = pd.concat([data, numbers], axis=0, ignore_index=True)
    # concatenate frequencies and masses to numbers
    numbers = pd.concat(
        [pd.DataFrame(freqs, columns=["freq"]), pd.DataFrame(masses, columns=["mass"]), numbers],
        axis=1,
    )
    numbers.to_csv(f"Magnitudes_{temp:.0f}K_.lx", index=False)
    for n in range(np.shape(structures)[2]):
        final_geom = structures[:, :, n]
        write_input(
            atomos,
            final_geom,
            header.replace("UUUUU", str(n + 1)),
            bottom.replace("UUUUU", str(n + 1)),
            f"Geometries/Geometry-{n+1+counter}-.com",
        )
    print("\n\nDone! Ready to run.")


################################################################


##COLLECTS RESULTS##############################################
def gather_data(opc, tipo):
    files = [
        file
        for file in os.listdir("Geometries")
        if ".log" in file and "Geometr" in file
    ]
    files = [
        i for i in files if "Normal termination" in open("Geometries/" + i, "r",encoding="utf-8").read()
    ]
    files = sorted(files, key=lambda file: float(file.split("-")[1]))
    with open("Samples.lx", "w",encoding="utf-8") as f:
        for file in files:
            num = file.split("-")[1]
            broadening = opc
            numeros, energies, fs, scfs = [], [], [], []
            corrected, total_corrected = -1, -1
            with open("Geometries/" + file, "r",encoding="utf-8") as g:
                for line in g:
                    if "Excited State" in line:
                        line = line.split()
                        numeros.append(line[2])
                        energies.append(line[4])
                        fs.append(line[8][2:])
                    elif "Corrected transition energy" in line:
                        line = line.split()
                        corrected = line[4]
                    elif "Total energy after correction" in line:
                        line = line.split()
                        total_corrected = 27.2114 * float(line[5])
                    elif "SCF Done:" in line:
                        line = line.split()
                        scfs.append(27.2114 * float(line[4]))
                if len(numeros) > 0:
                    f.write(
                        "Geometry "
                        + num
                        + ":  Vertical transition (eV) Oscillator strength Broadening Factor (eV) \n"
                    )
                    if corrected != -1 and tipo == "abs":  # abspcm
                        f.write(
                            f"Excited State {numeros[0]}\t{corrected}\t{fs[0]}\t{broadening}\n"
                        )
                    elif corrected != -1 and tipo == "emi":  # emipcm
                        energy = total_corrected - scfs[-1]
                        f.write(
                            f"Excited State {numeros[0]}\t{energy:.3f}\t{fs[0]}\t{broadening}\n"
                        )
                    elif corrected == -1 and tipo == "emi":
                        f.write(
                            f"Excited State {numeros[0]}\t{energies[0]}\t{fs[0]}\t{broadening}\n"
                        )
                    else:
                        for i, energy in enumerate(energies):
                            f.write(
                                f"Excited State {numeros[i]}\t{energy}\t{fs[i]}\t{broadening}\n"
                            )
                    f.write("\n")


###############################################################


##NORMALIZED GAUSSIAN##########################################
def gauss(x, v, s):
    y = (1 / (np.sqrt(2 * np.pi) * s)) * np.exp(-0.5 * ((x - v) / s) ** 2)
    return y


###############################################################


##COMPUTES AVG TRANSITION DIPOLE MOMENT########################
def calc_tdm(osc_strength, vertical_energy):
    # Energy terms converted to J
    term = E_CHARGE * (HBAR_J**2) / vertical_energy
    dipoles = np.sqrt(3 * term * osc_strength / (2 * MASS_E))
    # Conversion in au
    dipoles *= 1.179474389e29
    return np.mean(dipoles)


###############################################################


##PREVENTS OVERWRITING#########################################
def naming(arquivo):
    new_arquivo = arquivo
    if arquivo in os.listdir("."):
        duplo = True
        vers = 2
        while duplo:
            new_arquivo = str(vers) + arquivo
            if new_arquivo in os.listdir("."):
                vers += 1
            else:
                duplo = False
    return new_arquivo


###############################################################


##CALCULATES FLUORESCENCE LIFETIME IN S########################
def calc_emi_rate(xd, yd, dyd):
    # Integrates the emission spectrum
    integrated_emission = np.trapz(yd, xd)
    taxa = (1 / HBAR_EV) * integrated_emission
    error = (1 / HBAR_EV) * np.sqrt(np.trapz((dyd**2), xd))
    return taxa, error


###############################################################


##COMPUTES SPECTRA#############################################
def spectra(tipo, num_ex, nr):
    if tipo == "abs":
        constante = (
            (np.pi * (E_CHARGE**2) * HBAR_EV)
            / (2 * nr * MASS_E * LIGHT_SPEED * EPSILON_0)
            * 10 ** (20)
        )
    elif tipo == "emi":
        constante = (
            (nr**2)
            * (E_CHARGE**2)
            / (2 * np.pi * HBAR_EV * MASS_E * (LIGHT_SPEED**3) * EPSILON_0)
        )
    vertical_energy, osc_strength, sigma = [], [], []
    number = 0
    with open("Samples.lx", "r",encoding="utf-8") as f:
        for line in f:
            if "Geometry" in line:
                number += 1
            elif "Excited State" in line and int(line.split()[2][:-1]) in num_ex:
                line = line.split()
                if float(line[3]) <= 0:
                    print("Ignoring geom with negative vertical transition!")
                    number -= 1
                else:
                    vertical_energy.append(float(line[3]))
                    osc_strength.append(float(line[4]))
                    sigma.append(float(line[5]))
    coms = start_counter()
    if len(vertical_energy) == 0 or len(osc_strength) == 0:
        lx.parser.fatal_error("You need to run steps 1 and 2 first! Goodbye!")
    elif len(vertical_energy) != coms * max(num_ex):
        print(
            "Number of log files is less than the number of inputs. Something is not right! Computing the spectrum anyway..."
        )
    vertical_energy = np.array(vertical_energy)
    osc_strength = np.array(osc_strength)
    sigma = np.array(sigma)
    if tipo == "abs":
        espectro = constante * osc_strength
    else:
        espectro = constante * (vertical_energy**2) * osc_strength
        tdm = calc_tdm(osc_strength, vertical_energy)
    left = max(min(vertical_energy) - 3 * max(sigma), 0.001)
    right = max(vertical_energy) + 3 * max(sigma)
    x = np.linspace(left, right, int((right - left) / 0.01))
    if tipo == "abs":
        arquivo = "cross_section.lx"
        primeira = "{:8s} {:8s} {:8s}\n".format(
            "#Energy(ev)", "cross_section(A^2)", "error"
        )
    else:
        arquivo = "differential_rate.lx"
        primeira = "{:4s} {:4s} {:4s} TDM={:.3f} au\n".format(
            "#Energy(ev)", "diff_rate", "error", tdm
        )
    arquivo = naming(arquivo)
    y = espectro[:, np.newaxis] * gauss(x, vertical_energy[:, np.newaxis], sigma[:, np.newaxis])
    mean_y = np.sum(y, axis=0) / number
    # Error estimate
    sigma = np.sqrt(np.sum((y - mean_y) ** 2, axis=0) / (number * (number - 1)))

    if tipo == "emi":
        # Emission rate calculations
        mean_rate, error_rate = calc_emi_rate(x, mean_y, sigma)
        segunda = f"# Total Rate S1 -> S0: {mean_rate:5.2e} +/- {error_rate:5.2e} s^-1\n"
    else:
        segunda = "# Absorption from State: S0\n"

    print(number, "geometries considered.")
    with open(arquivo, "w",encoding="utf-8") as f:
        f.write(primeira)
        f.write(segunda)
        for i in range(0, len(x)):
            text = f"{x[i]:.6f} {mean_y[i]:.6e} {sigma[i]:.6e}\n"
            f.write(text)
    print(f"Spectrum printed in the {arquivo} file")


###############################################################

##FETCHES  FILES###############################################
def fetch_file(frase, ends):
    for file in [i for i in os.listdir(".")]:
        for end in ends:
            if end in file:
                return file
    lx.parser.fatal_error(f"No {frase} file found. Goodbye!")

###############################################################


##RUNS TASK MANAGER############################################
def batch():
    script = fetch_file("batch.sh", ["batch.sh"])
    limite = input("Maximum number of jobs to be submitted simultaneously?\n")
    num = input("Number of calculations in each job?\n")
    gaussian = input("g16 or g09?\n")
    try:
        limite = int(limite)
        int(num)
    except ValueError:
        lx.parser.fatal_error("These must be integers. Goodbye!")
    with open("limit.lx", "w",encoding="utf-8") as limit_file:
        limit_file.write(str(limite))
    subprocess.Popen(
        ["nohup", "lx_batch_run", script, gaussian, num, "&"]
    )

###############################################################


##RUNS W TUNING################################################
def omega_tuning():
    geomlog = fetch_file("input or log", [".com", ".log"])
    base, _, nproc, mem, _, _ = lx.parser.busca_input(geomlog)
    if "IOP" in base.upper() and ("108" in base or "107" in base):
        base2 = base.split()
        for elem in base2:
            if "/" in elem:
                base = elem
                break
    omega1 = "0.1"
    passo = "0.025"
    relax = "y"
    print("This is the configuration taken from the file:\n")
    print(f"Functional/basis: {base}")
    print("%nproc=" + nproc)
    print("%mem=" + mem)
    print(f"Initial Omega: {omega1} bohr^-1")
    print(f"Step: {passo} bohr^-1")
    print("Optimize at each step: yes")
    change = input("Are you satisfied with these parameters? y or n?\n")
    if change == "n":
        base = default(
            base,
            f"Functional/basis is {base}. If ok, Enter. Otherwise, type functional/basis.\n",
        )
        nproc = default(nproc, f"nproc={nproc}. If ok, Enter. Otherwise, type it.\n")
        mem = default(mem, f"mem={mem}. If ok, Enter. Otherwise, type it.\n")
        omega1 = default(
            omega1,
            f"Initial omega is {omega1} bohr^-1. If ok, Enter. Otherwise, type it.\n",
        )
        passo = default(
            passo,
            f"Initial step is {passo} bohr^-1. If ok, Enter. Otherwise, type it.\n",
        )
        relax = default(
            relax, "Optimize at each step: yes. If ok, Enter. Otherwise, type n\n"
        )
    script = fetch_file("batch script", ["batch.sh"])
    gaussian = input("g16 or g09?\n")
    parallel = input("Minimize submission of jobs: y/n\n")
    with open("limit.lx", "w",encoding="utf-8") as f:
        f.write("10")
    subprocess.Popen(
        [
            "nohup",
            "lx_omega",
            geomlog,
            base,
            nproc,
            mem,
            omega1,
            passo,
            relax,
            script,
            gaussian,
            parallel,
            "&",
        ]
    )

###############################################################


##RUNS CONFORMATIONAL SEARCH###################################
def conformational():
    freqlog = fetch_file("frequency", [".log"])
    freqs, _ = lx.parser.pega_freq(freqlog)
    freqs_active = freqs[:40]
    temp = int(HBAR_EV * freqs_active[-1] / BOLTZ_EV)
    delta_temp = int(temp / 10)
    temp, delta_temp = str(temp), str(delta_temp)
    base, _, nproc, mem, _, _ = lx.parser.busca_input(freqlog)
    if base == '':
        base = 'pm6'
    print("This is the configuration taken from the file:\n")
    print(f"Functional/basis: {base}")
    print(f"%nproc={nproc}")
    print(f"%mem={mem}")
    print(f"Initial Temperature: {temp} K")
    print(f"Temperature step: {delta_temp} K")
    change = input("Are you satisfied with these parameters? y or n?\n")
    if change == "n":
        base = default(
            base,
            f"Functional/basis is {base}. If ok, Enter. Otherwise, type functional/basis.\n",
        )
        nproc = default(nproc, f"nproc={nproc}. If ok, Enter. Otherwise, type it.\n")
        mem = default(mem, f"mem={mem}. If ok, Enter. Otherwise, type it.\n")
        temp = default(
            temp, f"Initial temperature is {temp} K. If ok, Enter. Otherwise, type it.\n"
        )
        delta_temp = default(
            delta_temp, f"Temperature step is {delta_temp} K. If ok, Enter. Otherwise, type it.\n"
        )
    script = fetch_file("batch script", ["batch.sh"])
    num_geoms = input("Number of geometries sampled at each round?\n")
    rounds = input("Number of rounds?\n")
    numjobs = input("Number of jobs in each batch?\n")
    gaussian = input("g16 or g09?\n")
    try:
        int(num_geoms)
        int(rounds)
        int(numjobs)
    except ValueError:
        lx.parser.fatal_error("These must be integers. Goodbye!")
    with open("limit.lx", "w",encoding="utf-8") as f:
        f.write("10")
    subprocess.Popen(
        [
            "nohup",
            "lx_conf_search",
            freqlog,
            base,
            nproc,
            mem,
            temp,
            delta_temp,
            num_geoms,
            rounds,
            numjobs,
            script,
            gaussian,
            "&",
        ]
    )


###############################################################


##FINDS SUITABLE VALUE FOR STD#################################
def detect_sigma():
    try:
        files = [i for i in os.listdir(".") if "Magnitudes" in i and ".lx" in i]
        file = files[0]
        temp = float(file.split("_")[1].strip("K"))
        sigma = np.round(BOLTZ_EV * temp, 3)
    except (FileNotFoundError, IndexError):
        sigma = 0.000
    return sigma


###############################################################


##CHECKS SPECTRUM TYPE#########################################
def get_spec():
    coms = [
        file
        for file in os.listdir("Geometries")
        if "Geometr" in file and ".com" in file
    ]
    with open("Geometries/" + coms[0], "r",encoding="utf-8") as f:
        for line in f:
            if "ABSSPCT" in line:
                tipo = "absorption"
                break
            elif "EMISPCT" in line:
                tipo = "emission"
                break
    return tipo


###############################################################


##QUERY FUNCTION###############################################
def default(a, frase):
    b = input(frase)
    if b == "":
        return a
    return b


###############################################################


##SETS DIELECTRIC CONSTANTS####################################
def set_eps(scrf):
    if "READ" in scrf.upper():
        eps1 = input("Type the static dielectric constant.\n")
        eps2 = input("Type the dynamic dielectric constant (n^2).\n")
        try:
            float(eps1)
            float(eps2)
        except ValueError:
            lx.parser.fatal_error("The constants must be numbers. Goodbye!")
        epss = "Eps=" + eps1 + "\nEpsInf=" + eps2 + "\n\n"
    else:
        epss = "\n"
    return epss


###############################################################


##STOP SUBMISSION OF JOBS######################################
def abort_batch():
    choice = input(
        "Are you sure you want to prevent new jobs from being submitted? y or n?\n"
    )
    if choice == "y":
        try:
            os.remove("limit.lx")
            print("Done!")
        except FileNotFoundError:
            print("Could not find the files. Maybe you are in the wrong folder.")
    else:
        print("OK, nevermind")


###############################################################


##DELETES CHK FILES############################################
def delchk(input_file):
    num = input_file.split("-")[1]
    chks = [i for i in os.listdir(".") if f"_{num}.chk" in i]
    for chk in chks:
        try:
            os.remove(chk)
        except FileNotFoundError:
            pass


###############################################################

##CHECKS WHETHER THE JOB IS TWO STEP###########################
def set_factor(file):
    factor = 1
    with open(file, "r",encoding="utf-8") as f:
        for line in f:
            if "Link1" in line:
                factor = 2
                break
    return factor
###############################################################

class Watcher:
    def __init__(self, folder,files=None,counter=1):
        self.folder = folder
        if files is None:
            self.files = [i[:-4] for i in os.listdir(folder) if i.endswith('.com') and "Geometr" in i]
            self.files = sorted(self.files, key=lambda pair: float(pair.split("-")[1]))
        else:
            self.files = [i[:-4] for i in files]
        self.number_inputs = len(self.files)
        self.done = []
        self.error = []
        self.running = []
        self.running_batches = 0
        self.counter = counter

    def check(self):
        list_to_check = self.files.copy()
        for input_file in list_to_check:
            term = 0
            try:
                with open(self.folder + "/" + input_file + ".log", "r",encoding="utf-8") as log_file:
                    for line in log_file:
                        if "Normal termination" in line:
                            term += 1
                            if term == self.counter:
                                if self.counter == 2:
                                    delchk(input)
                                self.done.append(input_file)
                                del self.files[self.files.index(input_file)]
                                break
                        elif "Error termination" in line:
                            self.error.append(input_file)
                            del self.files[self.files.index(input_file)]
                            break
            except FileNotFoundError:
                pass

    def report(self):
        self.check()
        print('\n\n')
        print(f'There are {len(self.done)} successfully completed calculations out of {self.number_inputs} inputs.')
        print(f'{100 * len(self.done) / self.number_inputs:.1f}% of the calculations have been run.')
        if len(self.error) > 0:
            print(f"There are {len(self.error)} failed jobs.")
            print('These are: ', self.error)

    def limit(self):
        if self.folder == ".":
            fold = "."
        else:
            fold = '..'
        try:
            return np.loadtxt(f"{fold}/limit.lx",encoding='utf-8')
        except (OSError,FileNotFoundError):
            sys.exit()

    def keep_going(self,num):
        if len(self.running) / num < self.limit():
            return False
        return True

    def clean_failed(self):
        for failed in self.error:
            os.remove(self.folder + "/" + failed + ".log")
        self.files += self.error
        self.error = []

    def run(self, batch_file, gaussian, num):
        self.check()
        self.clean_failed()
        inputs = self.files.copy()
        while len(inputs) > 0:
            next_inputs = inputs[:int(num)]
            command = ''
            for input_file in next_inputs:
                command += f"{gaussian} {input_file}.com \n"
                self.running.append(input_file)
                inputs.remove(input_file)
            command += "wait"
            with open(f"cmd_{self.running_batches}_.sh", "w",encoding='utf-8') as cmd:
                cmd.write(command)
            sub = subprocess.call(["bash", batch_file, f"cmd_{self.running_batches}_.sh"])
            self.running_batches += 1
            keep = self.keep_going(num)
            while keep:
                time.sleep(20)
                self.check()
                concluded = self.done + self.error
                self.running = [elem for elem in self.running if elem not in concluded]
                keep = self.keep_going(num)

    def hold_watch(self):
        while len(self.files) > 0:
            self.check()
            _ = self.limit()
            time.sleep(20)

###############################################################

##CHECKS PROGRESS##############################################
def andamento():
    files = [i for i in os.listdir("Geometries") if "Geometr" in i and ".com" in i]
    factor = set_factor(files[0])
    the_watcher = Watcher('Geometries',counter=factor)
    the_watcher.report()
###############################################################

##GETS SPECTRA#################################################
def search_spectra():
    absorption, emission = "None", "None"
    candidates = [i for i in os.listdir(".") if ".lx" in i]
    for candidate in candidates:
        with open(candidate, "r",encoding="utf-8") as f:
            for line in f:
                if "cross_section" in line:
                    absorption = candidate
                elif "diff_rate" in line:
                    emission = candidate
                break
    return absorption, emission


###############################################################

##RUNS EXCITON ANALYSIS########################################
def ld():
    absorption, emission = search_spectra()
    print(f"Absorption file: {absorption}")
    print(f"Emission file: {emission}")
    check = input("Are these correct? y or n?\n")
    if check == "n":
        absorption = input("Type name of the absorption spectrum file\n")
        emission = input("Type name of the emission spectrum file\n")

    kappa = input("Orientation Factor (k^2):\n")
    rmin = input("Average intermolecular distance in Å:\n")
    quantum_yield = input("Fluorescence quantum yield (from 0 to 1):\n")
    try:
        rmin = float(rmin)
        kappa = np.sqrt(float(kappa))
        quantum_yield = float(quantum_yield)
    except ValueError:
        lx.parser.fatal_error("These features must be numbers. Goodbye!")
    if quantum_yield > 1 or quantum_yield < 0:
        lx.parser.fatal_error("Quantum yield must be between 0 and 1. Goodbye!")

    correct = input("Include correction for short distances? y or n?\n")
    if correct == "y":
        alpha = 1.15 * 0.53
        print("Employing correction!")
    else:
        alpha = 0
        print("Not employing correction!")

    print("Computing...")
    try:
        run_ld(absorption, emission, alpha, rmin, kappa, quantum_yield)
        print("Results can be found in the ld.lx file")
    except Exception as e:
        print("Something went wrong. Check if the name of the files are correct.")
        print(e)


###############################################################

def check_for_updates(package_name):
    try:
        # Get the currently installed version
        installed_version = pkg_resources.get_distribution(package_name).version
        
        # Fetch the latest version from PyPI
        response = requests.get(f'https://pypi.org/pypi/{package_name}/json')
        response.raise_for_status()
        latest_version = response.json()['info']['version']

        # Compare versions
        if installed_version != latest_version:
            print(f"ATTENTION: Update available! {package_name} {installed_version} -> {latest_version}")
            print("Run `pip install --upgrade {}` to update.".format(package_name))

    except Exception as e:
        print(f"An error occurred while checking for updates: {e}")


#!/usr/bin/env python3
import sys
import lx.tools
import lx.parser
from lx.conf_search import classify_only
from lx import __version__ as lx_version

def interface():
    print(" ▄█          ▄████████  ▄██████▄  ▀████    ▐████▀")
    print("███         ███    ███ ███    ███   ███▌   ████▀ ")
    print("███         ███    █▀  ███    ███    ███  ▐███   ")
    print("███        ▄███▄▄▄     ███    ███    ▀███▄███▀   ")
    print("███       ▀▀███▀▀▀     ███    ███    ████▀██▄    ")
    print("███         ███    █▄  ███    ███   ▐███  ▀███   ")
    print("███▌    ▄   ███    ███ ███    ███  ▄███     ███▄ ")
    print("█████▄▄██   ██████████  ▀██████▀  ████       ███▄")
    print("▀                                                ")
    print(f"Version: {lx_version.__version__}\n")
    print("Choose your option:\n")
    print("SPECTRUM SIMULATIONS:")
    print("\t1 - Generate the inputs for the spectrum calculation")
    print("\t2 - Run the spectrum calculations")
    print("\t3 - Generate the spectrum")
    print("\t4 - Check the progress of the calculations")
    print("EXCITON ANALYSIS:")
    print(
        "\t5 - Estimate Förster radius, fluorescence lifetime and exciton diffusion lengths"
    )
    print("CONFORMATIONAL ANALYSIS:")
    print("\t6 - Perform conformational search")
    print("OTHER FEATURES:")
    print("\t7 - Perform long-range parameter tuning")
    print("\t8 - Get rid of imaginary frequencies")
    print("\t9 - Retrieve last geometry from log file")
    print("\t10 - Abort my calculations")
    print('\n')
    lx.tools.check_for_updates('LeoX')
    op = input()
    if op == "1":
        freqlog = lx.tools.fetch_file("frequency", [".log"])
        freqs, _ = lx.parser.pega_freq(freqlog)
        if freqs[0] < 0:
            lx.parser.fatal_error("Imaginary frequency! Goodbye!")
        cm = lx.parser.get_cm(freqlog)
        base, temtd, nproc, mem, scrf, spec = lx.parser.busca_input(freqlog)
        if temtd == "":
            tda = "TD"
        else:
            tda = temtd.upper()
        print("\nThe suggested configurations for you are:\n")
        print(f"Functional/basis: {base}")
        print(f"Solvent Method: {scrf}")
        print(f"Charge and Multiplicity: {cm}")
        print(f"TD-DFT or TDA-DFT: {tda}-DFT")
        print("%nproc=" + nproc)
        print("%mem=" + mem)
        change = input("Are you satisfied with these parameters? y or n?\n")
        if change.lower() == "n":
            base = lx.tools.default(
                base,
                f"Functional/basis is {base}. If ok, Enter. Otherwise, type functional/basis.\n",
            )
            scrf = lx.tools.default(
                scrf,
                f"SCRF keyword is {scrf}. If ok, Enter. Otherwise, type the desired one.\n",
            )
            cm = lx.tools.default(
                cm,
                f"Charge and multiplicity is {cm}. If ok, Enter. Otherwise, type charge and multiplicity Ex.: 0 1\n",
            )
            nproc = lx.tools.default(
                nproc, f"nproc is {nproc}. If ok, Enter. Otherwise, type it.\n"
            )
            mem = lx.tools.default(
                mem, f"mem is {mem}. If ok, Enter. Otherwise, type it.\n"
            )
            tamm = input("Use TDA (Tamm-Dancoff Approximation)? y or n?\n")
            if tamm.lower() == "y":
                tda = "TDA"
            else:
                tda = "TD"

        num_ex = input("How many excited states?\n")
        try:
            num_ex = int(num_ex)
        except ValueError:
            lx.parser.fatal_error("This must be a number! Goodbye!")
        num_geoms = int(input("How many geometries to be sampled?\n"))
        pcm = input("Include state specific solvent approach? y or n?\n")
        if pcm.lower() == "y":
            solv = input(
                "What is the solvent? If you want to specify the dielectric constants yourself, type read.\n"
            )
            epss = lx.tools.set_eps(solv)
            if epss == "\n":
                solv = "SOLVENT=" + solv
            if temtd:
                print("Inputs suitable for emission spectra!\n")
                header = f"%chk=step_UUUUU.chk\n%nproc={nproc}\n%mem={mem}\n# {base} {tda}=(NSTATES={num_ex}) SCRF=(CorrectedLR,NonEquilibrium=Save,{solv})\n\n{spec}\n\n{cm}\n"
                bottom = f"{epss}\n--Link1--\n%nproc={nproc}\n%mem={mem}\n%oldchk=step_UUUUU.chk\n%chk=step2_UUUUU.chk\n# {base} GUESS=READ GEOM=CHECKPOINT SCRF(NonEquilibrium=Read,{solv})\n\nTITLE\n\n{cm}\n\n{epss}"
            else:
                print("Inputs suitable for absortion spectra!!\n")
                header = f"%nproc={nproc}\n%mem={mem}\n# {base} SCRF=(CorrectedLR,{solv}) {tda}=(NSTATES={num_ex})\n\n{spec}\n\n{cm}\n"
                bottom = epss
        elif pcm == "n":
            epss = lx.tools.set_eps(scrf)
            header = f"%nproc={nproc}\n%Mem={mem}\n# {tda}=(NStates={num_ex}) {base} {scrf} \n\n{spec}\n\n{cm}\n"
            bottom = epss + "\n\n"
        else:
            lx.parser.fatal_error("It should be y or n. Goodbye!")
        temperature = float(input("Temperature in Kelvin?\n"))
        if temperature <= 0:
            lx.parser.fatal_error("Have you heard about absolute zero? Goodbye!")
        lx.tools.make_ensemble(freqlog, num_geoms, temperature, header, bottom)
    elif op == "2":
        lx.tools.batch()
    elif op == "3":
        opc = lx.tools.detect_sigma()
        tipo = lx.tools.get_spec()
        nr = lx.parser.get_nr()
        print("The spectrum will be run with the following parameters:\n")
        print(f"Spectrum type: {tipo.title()}")
        print(f"Standard deviation of: {opc:.3f} eV")
        print(f"Refractive index: {nr:.3f}\n")
        change = input("Are you satisfied with these parameters? y or n?\n")
        if change.lower() == "n":
            opc = input("What is the standard deviation of the gaussians?\n")
            try:
                opc = float(opc)
            except ValueError:
                lx.parser.fatal_error("It must be a number. Goodbye!")
            nr = input("What is the refractive index?\n")
            try:
                nr = float(nr)
            except ValueError:
                lx.parser.fatal_error("It must be a number. Goodbye!")
            tipo = input(
                "What kind of spectrum? Type abs (absorption) or emi (emission)\n"
            )
            if tipo != "abs" and tipo != "emi":
                lx.parser.fatal_error("It must be either one. Goodbye!")
        tipo = tipo[:3]
        if tipo == "abs":
            estados = input("How many excited states?\n")
            try:
                estados = int(estados)
            except ValueError:
                lx.parser.fatal_error("It must be an integer! Goodbye!")
        else:
            estados = 1
        num_ex = range(0, estados + 1)
        num_ex = list(map(int, num_ex))
        lx.tools.gather_data(opc, tipo)
        lx.tools.spectra(tipo, num_ex, nr)
    elif op == "4":
        lx.tools.andamento()
    elif op == "5":
        lx.tools.ld()
    elif op == "6":
        question = input("Classify only? y or n?\n")
        if question.lower() == "y":
            try:
                classify_only()
            except:
                lx.parser.fatal_error(
                    "Something went wrong. Your folder may not contain Geometry log files. Goodbye!"
                )
        else:
            lx.tools.conformational()
    elif op == "7":
        lx.tools.omega_tuning()
    elif op == "8":
        freqlog = lx.tools.fetch_file(
            "Frequency log with imaginary frequencies", [".log"]
        )
        lx.tools.distort(freqlog)
    elif op == "9":
        freqlog = lx.tools.fetch_file("log", [".log"])
        base, _, nproc, mem, scrf, _ = lx.parser.busca_input(freqlog)
        cm = lx.parser.get_cm(freqlog)
        header = f"%nproc={nproc}\n%mem={mem}\n# {base} {scrf}\n\nTITLE\n\n{cm}\n"
        geom, atomos = lx.parser.pega_geom(freqlog)
        lx.tools.write_input(atomos, geom, header, "", "geom.lx")
        print("Geometry saved in the geom.lx file.")
    elif op == "10":
        lx.tools.abort_batch()
    else:
        lx.parser.fatal_error("It must be one of the options... Goodbye!")


def main():
    try:
        freqlog = sys.argv[1]
        geom, atomos = lx.parser.pega_geom(freqlog)
        print(len(atomos))
        print("")
        for i, atomo in enumerate(atomos):
            print(f"{atomo:2s}  {geom[i,0]:.7f}  {geom[i,1]:.7f}  {geom[i,2]:.7f}")
    except IndexError:
        interface()


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
import numpy as np
from scipy.interpolate import interp1d
import lx.parser

##SOME CONSTANTS##############################################
EPSILON_0 = lx.parser.EPSILON_0  # F/m
HBAR_EV = lx.parser.HBAR_EV  # eV s
HBAR_J = lx.parser.HBAR_J  # J s
MASS_E = lx.parser.MASS_E  # kg
LIGHT_SPEED = lx.parser.LIGHT_SPEED  # m/s
E_CHARGE = lx.parser.E_CHARGE  # C
BOLTZ_EV = lx.parser.BOLTZ_EV  # eV/K
AMU = lx.parser.AMU  # kg
###############################################################


##CALCULATES FLUORESCENCE LIFETIME IN S########################
def calc_lifetime(xd, yd, dyd):
    # Integrates the emission spectrum
    IntEmi = np.trapz(yd, xd)
    taxa = (1 / HBAR_EV) * IntEmi
    error = (1 / HBAR_EV) * np.sqrt(np.trapz((dyd**2), xd))
    dlife = (1 / taxa) * (error / taxa)
    return 1 / taxa, dlife


###############################################################


##CALCULATES FORSTER RADIUS####################################
def radius(xa, ya, dya, xd, yd, dyd, kappa):
    # Finds the edges of interpolation
    minA = min(xa)
    minD = min(xd)
    maxA = max(xa)
    maxD = max(xd)
    MIN = max(minA, minD)
    MAX = min(maxA, maxD)

    if MIN > MAX:
        return 0, 0
    X = np.linspace(MIN, MAX, 1000)
    f1 = interp1d(xa, ya, kind="cubic")
    f2 = interp1d(xd, yd, kind="cubic")
    f3 = interp1d(xa, dya, kind="cubic")
    f4 = interp1d(xd, dyd, kind="cubic")

    YA = f1(X)
    YD = f2(X)
    DYA = f3(X)
    DYD = f4(X)

    # Calculates the overlap
    Overlap = YA * YD / (X**4)

    # Overlap error
    OverError = Overlap * np.sqrt((DYA / YA) ** 2 + (DYD / YD) ** 2)

    # Integrates overlap
    IntOver = np.trapz(Overlap, X)

    # Integrated Overlap Error
    DeltaOver = np.sqrt(np.trapz((OverError**2), X))

    # Gets lifetime
    tau, delta_tau = calc_lifetime(xd, yd, dyd)

    # Calculates radius sixth power
    c = LIGHT_SPEED * 1e10
    const = (HBAR_EV**3) * (9 * (c**4) * (kappa**2) * tau) / (8 * np.pi)
    radius6 = const * IntOver

    # Relative error in radius6
    delta = np.sqrt((DeltaOver / IntOver) ** 2 + (delta_tau / tau) ** 2)

    # Calculates radius
    forster_radius = radius6 ** (1 / 6)

    # Error in radius
    delta_radius = forster_radius * delta / 6

    return forster_radius, delta_radius


###############################################################


##AVG HOPPING DISTANCES########################################
def r_avg(alpha, moment, rmin, dim):
    neighbors = 100
    n1 = neighbors
    n2 = 0
    n3 = 0
    if dim > 1:
        n2 = neighbors
    if dim > 2:
        n3 = neighbors
    numerator = []
    denominator = []
    for i in range(0, n1 + 1):
        for j in range(0, n2 + 1):
            for k in range(0, n3 + 1):
                r = rmin * np.sqrt(i**2 + j**2 + k**2)
                if r != 0:
                    numerator.append(r / ((alpha * moment + r) ** 6))
                    denominator.append(1.0 / ((alpha * moment + r) ** 6))
    average = np.sum(numerator) / np.sum(denominator)
    return average


###############################################################


##1D PROJECTIONS OF DIFFUSION LENGTH###########################
def difflen(forster_radius, alpha, moment, r, dim, phi):
    ld = r * (forster_radius**3) / ((alpha * moment + r) ** 3)
    ld = np.sqrt(phi) * ld / np.sqrt(dim)
    # Conversion to nm
    ld = ld / 10
    return ld


######## ANNIHILATION COEFs ###################################
def KSSA(RF, r, tau, error_radius, error_life):
    KSSA_cte = (4 * np.pi / tau) * (
        (RF) ** 6 / (r**3)
    )  # Singlet-Singlet annihilation coef
    RF_6_power_error = np.sqrt(6 * ((error_radius / RF) ** 2))
    delta = np.sqrt(RF_6_power_error**2 + (error_life / tau) ** 2)
    KSSA_error = KSSA_cte * delta
    un = 1e-24  # unit factor 1 AA^3 = 1E-24 cm^3
    KSSA_cte, KSSA_error = KSSA_cte * un, KSSA_error * un
    return KSSA_cte, KSSA_error


def KTTA(RF, r, tau, error_radius, error_life):
    KTTA_cte = (4 * np.pi / tau) * ((RF) ** 6)  # Triplet-Triplet annihilation coef
    RF_6_power_error = np.sqrt(6 * ((error_radius / RF) ** 2))
    delta = np.sqrt(RF_6_power_error**2 + (error_life / tau) ** 2)
    KTTA_error = KTTA_cte * delta
    un = 1e-48  # unit factor 1 AA^6 = 1E-48 cm^6
    KTTA_cte, KTTA_error = KTTA_cte * un, KTTA_error * un
    return KTTA_cte, KTTA_error


###############################################################


##RETURNS ABS AND EMISSION TYPES###############################
def emi_abs_types(Abs, Emi):
    abs_type = "S0"
    with open(Emi, "r",encoding='utf-8') as f:
        line = f.readlines()[1].split("->")
        emi_init = line[0].split()[-1].strip()
        emi_final = line[1].split(":")[0].strip()
    return abs_type, emi_init, emi_final


###############################################################


##Calculates average hopping distances and diffusion length for each case ####
def get_lds_dists(alpha, moment, rmin, Phi, mean_radius, error_radius):
    # convention: the last list element (with dimension 3) refers to the Amorphous one
    # dists' structure: [average distance,dimension,morphology label]
    # lds'   structure: [ld , error ld]
    dists = [
        [r_avg(alpha, moment, rmin, i), i, "%sD Crystal" % (i)] for i in range(1, 4)
    ]  # Calculates average hopping distances for each case
    dists.append([0.25 * alpha * moment + 1.25 * rmin, 3, "Amorphous "])
    lds = [
        [
            difflen(mean_radius, alpha, moment, dists[i][0], dists[i][1], Phi),
            3
            * difflen(mean_radius, alpha, moment, dists[i][0], dists[i][1], Phi)
            * (error_radius / mean_radius),
        ]
        for i in range(len(dists))
    ]
    return lds, dists


###############################################################


def run_ld(Abs, Emi, alpha, rmin, kappa, Phi):
    # Gets the Transition Dipole Moment in atomic units
    try:
        with open(Emi, "r",encoding='utf-8') as f:
            for line in f:
                moment = float(line.split("=")[1].split()[0])
                break
    except (FileNotFoundError,IndexError, ValueError):
        moment = 0

    # Loads the data
    data_abs = np.loadtxt(Abs)
    data_emi = np.loadtxt(Emi)

    # getting the types of abs and emi spectras
    abs_type, emi_init, emi_final = emi_abs_types(Abs, Emi)

    # template : abs = S0, emi = Sn ---> S0 ===> ('S0','S','S0')
    rates_types = {
        ("S0", "S", "S0"): "SSA",
        ("T1", "T", "S0"): "TTA",
        ("S0", "T", "S0"): "Only radius",
        ("S1", "S", "S0"): "Only radius",
    }

    try:
        setting_spectras = rates_types[(abs_type, emi_init[0], emi_final)]
    except KeyError:
        print(
            f"The configuration:\nabs {abs_type}\nemi {emi_init} -> {emi_final}\nis not defined! Generating standard calculations anyway ..."
        )
        setting_spectras = "Only radius"

    # Gets energies, intensities and errors for the donor and acceptor
    xa, ya, dya = data_abs[:, 0], data_abs[:, 1], data_abs[:, 2]
    xd, yd, dyd = data_emi[:, 0], data_emi[:, 1], data_emi[:, 2]

    # Lifetime calculations
    mean_life, error_life = calc_lifetime(xd, yd, dyd)

    # Radius calculations
    mean_radius, error_radius = radius(xa, ya, dya, xd, yd, dyd, kappa)

    # Diffusion length and distance calculations for all conformations
    lds, dists = get_lds_dists(alpha, moment, rmin, Phi, mean_radius, error_radius)

    with open("ld.lx", "w",encoding='utf-8') as f:
        f.write(f"Abs: {abs_type}, Emi: {emi_init} -> {emi_final}\n")
        f.write(
            f"Forster Radius:      {mean_radius:.1f} +/- {error_radius:.1f} AA \n"
        )
        f.write(
            f"Radiative Lifetime:  {mean_life:.3e} +/- {error_life:.3e} s\n"
        )
        f.write(f"Avg. Dipole Moment:  {moment:.1f} a.u. \n")

        if setting_spectras == "SSA":
            KSSA_cte, KSSA_error = KSSA(
                mean_radius, rmin, mean_life, error_radius, error_life
            )
            f.write(
                f"Annihilation Coefficient Singlet-Singlet: {KSSA_cte:5.2e} +/- {KSSA_error:5.2e} cm^3 s^-1\n"
            )
            f.write("Morphology   Avg_Hop_Distance(AA)  Diffusion_Length(nm)\n")
            for i, dist in enumerate(dists):
                f.write(f"{dist[2]}   {dist[0]:<19.1f}  {lds[i][0]:>.1f} +/- {lds[i][1]:>.1f}\n")
        if setting_spectras == "TTA":
            KTTA_cte, KTTA_error = KTTA(
                mean_radius, rmin, mean_life, error_radius, error_life
            )
            f.write(
                f"Annihilation Coefficient Triplet-Triplet: {KTTA_cte:5.2e} +/- {KTTA_error:5.2e} cm^6 s^-1\n"
            )
        if setting_spectras == "Only radius":
            pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev

import io
import os
import sys
import subprocess
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = 'LeoX'
DESCRIPTION = 'Spectrum simulations with TD(A)-DFT'
URL = 'https://github.com/LeonardoESousa/LeoX'
EMAIL = 'leonardo.sousa137@gmail.com'
AUTHOR = 'Leonardo Evaristo de Sousa'
REQUIRES_PYTHON = '>=3.6.0'

# What packages are required for this module to be executed?
REQUIRED = ['numpy', 'scipy', 'pandas','wheel']

# What packages are optional?
EXTRAS = {
    # 'fancy feature': ['django'],
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
project_slug = 'lx'
with open(os.path.join(here, project_slug, '__version__.py'),encoding='utf-8') as f:
    exec(f.read(), about)

class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print(f'\033[1m{s}\033[0m')

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        subprocess.run([sys.executable, 'setup.py', 'sdist', 'bdist_wheel', '--universal'], check=True)

        self.status('Uploading the package to PyPI via Twine…')
        subprocess.run(['twine', 'upload', 'dist/*'], check=True)

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    # If your package is a single module, use this instead of 'packages':
    #py_modules=['mypackage'],
    entry_points={'console_scripts': ["lx = lx.__main__:main", "lx_batch_run = lx.batch:run_batch", "lx_omega = lx.omega:main", "lx_conf_search = lx.conf_search:main"]},
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license='MIT',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
    },
)



import sys
import os
import numpy as np

##SOME CONSTANTS##############################################
EPSILON_0 = 8.854187817e-12  # F/m
HBAR_EV = 6.582119514e-16  # eV s
HBAR_J = 1.054571800e-34  # J s
MASS_E = 9.10938356e-31  # kg
LIGHT_SPEED = 299792458  # m/s
E_CHARGE = 1.60217662e-19  # C
BOLTZ_EV = 8.6173303e-5  # eV/K
AMU = 1.660539040e-27  # kg
###############################################################


##ERROR FUNCTION###############################################
def fatal_error(msg):
    print(msg)
    sys.exit()


###############################################################


##GETS FREQUENCIES AND REDUCED MASSES##########################
def pega_freq(freqlog):
    freqs, masses = [], []
    with open(freqlog, "r",encoding='utf-8') as f:
        for line in f:
            if "Frequencies -- " in line:
                line = line.split()
                for j in range(2, len(line)):
                    freqs.append(float(line[j]))
            elif "Red. masses --" in line:
                line = line.split()
                for j in range(3, len(line)):
                    masses.append(float(line[j]))
            elif "Thermochemistry" in line:
                break
    if len(freqs) == 0 or len(masses) == 0:
        fatal_error(
            "No frequencies or reduced masses found. Check your frequency file. Goodbye."
        )
    # conversion in angular frequency
    freqs = np.array(freqs) * (LIGHT_SPEED * 100 * 2 * np.pi)
    # conversion from amu to kg
    masses = np.asarray(masses) * AMU
    return freqs, masses


###############################################################


def pega_geom(freqlog):
    if ".log" in freqlog:
        busca = " orientation:"
        fetch = False
        with open(freqlog, "r",encoding='utf-8') as f:
            for line in f:
                if busca in line and "Dipole" not in line:
                    geom = np.zeros((1, 3))
                    atomos = []
                    fetch = True
                elif fetch:
                    line = line.split()
                    try:
                        float(line[0])
                        NG = []
                        for j in range(3, len(line)):
                            NG.append(float(line[j]))
                        atomos.append(line[1])
                        geom = np.vstack((geom, NG))
                    except ValueError:
                        if len(atomos) > 0:
                            fetch = False
    else:
        geom = np.zeros((1, 3))
        atomos = []
        with open(freqlog, "r",encoding='utf-8') as f:
            for line in f:
                line = line.split()
                try:
                    vetor = np.array([float(line[1]), float(line[2]), float(line[3])])
                    atomos.append(line[0])
                    geom = np.vstack((geom, vetor))
                except (IndexError, ValueError):
                    pass
    geom = geom[1:, :]
    return geom, atomos


###############################################################


##GETS NORMAL COORDINATES IN HIGH PRECISION####################
def pega_modosHP(G, freqlog):
    F, _ = pega_freq(freqlog)
    num_atoms = np.shape(G)[0]
    normal_modes = np.zeros((num_atoms, 3, len(F))) + np.nan
    mode = 0
    fetch = False
    with open(freqlog, "r",encoding='utf-8') as f:
        for line in f:
            if "Coord Atom Element:" in line:
                fetch = True
            elif fetch:
                line = line.split()
                if len(line) < 6 or "Harmonic" in line[0]:
                    fetch = False
                    mode += 5
                else:
                    coord = int(line[0]) - 1
                    atom = int(line[1]) - 1
                    for j in range(3, len(line)):
                        normal_modes[atom, coord, mode + j - 3] = float(line[j])
    return normal_modes


###############################################################


##GETS NORMAL COORDINATES IN REGULAR PRECISION#################
def pega_modosLP(G, freqlog):
    F, _ = pega_freq(freqlog)
    num_atoms = np.shape(G)[0]
    normal_modes = np.zeros((num_atoms, 3, len(F))) + np.nan
    mode = 0
    fetch = False
    with open(freqlog, "r",encoding='utf-8') as f:
        for line in f:
            if "Atom  AN      X      Y      Z" in line:
                fetch = True
            elif fetch:
                line = line.split()
                if len(line) < 5:
                    fetch = False
                    mode += 3
                else:
                    atom = int(line[0]) - 1
                    for j in range(2, len(line)):
                        normal_modes[atom, (j - 2) % 3, mode + (j - 2) // 3] = float(
                            line[j]
                        )
    return normal_modes


###############################################################


##DETECTS WHETHER HIGH PRECISION IS USED#######################
def pega_modos(geom, freqlog):
    x = "LP"
    with open(freqlog, "r",encoding='utf-8') as f:
        for line in f:
            if "Coord Atom Element:" in line:
                x = "HP"
                break
    if x == "LP":
        return pega_modosLP(geom, freqlog)
    else:
        return pega_modosHP(geom, freqlog)


###############################################################


##GETS INPUT PARAMS FROM LOG FILES#############################
def get_input_params(freqlog):
    nproc, mem, header = "", "", ""
    with open(freqlog, "r",encoding='utf-8') as f:
        search = False
        for line in f:
            if "%nproc" in line.lower():
                line = line.split("=")
                nproc = line[-1].replace("\n", "")
            elif "%mem" in line.lower():
                line = line.split("=")
                mem = line[-1].replace("\n", "")
            elif "#" in line and not search and header == "":
                search = True
                header += line.lstrip().replace("\n", "")
            elif search and "----------" not in line:
                header += line.lstrip().replace("\n", "")
            elif search and "----------" in line:
                search = False
                break
    return nproc, mem, header


###############################################################


##CHECKS THE FREQUENCY LOG'S LEVEL OF THEORY###################
def busca_input(freqlog):
    base = "lalala"
    exc = ""
    header = ""
    nproc = "4"
    mem = "1GB"
    scrf = ""
    nproc, mem, header = get_input_params(freqlog)

    if "TDA" in header.upper():
        exc = "tda"
        spec = "EMISPCT"
    elif "TD" in header.upper():
        exc = "td"
        spec = "EMISPCT"
    else:
        spec = "ABSSPCT"

    if "SCRF" in header.upper():
        new = header.split()
        for elem in new:
            if "SCRF" in elem:
                scrf = elem
                break

    header = header.split()
    base = ""
    for elem in header:
        if "/" in elem and "IOP" not in elem.upper():
            base += elem.replace("#", "")
        elif "IOP" in elem.upper() and ("108" in elem or "107" in elem):
            base += " " + elem
    return base, exc, nproc, mem, scrf, spec


###############################################################


##CHECKS FREQ FILES############################################
def double_check(freqlog):
    freqs, _ = pega_freq(freqlog)
    if np.any(freqs < 0):
        fatal_error(
            "Imaginary frequencies detected. Check your frequency file. Goodbye."
        )
    _, _, header = get_input_params(freqlog)
    header = header.lower()
    optfreqissue = False
    stationary = False
    if "opt" in header and "freq" in header and "iop(" in header:
        optfreqissue = True
    with open(freqlog, "r",encoding='utf-8') as f:
        for line in f:
            if "Stationary point found" in line:
                stationary = True
            elif all(s in line for s in ["Item", "Value", "Threshold", "Converged?"]):
                stationary = False
    if not stationary:
        print("*" * 50)
        print("WARNING: Non-optimized parameters detected in your frequency file.")
        print(
            "Even though the frequencies may be all real, your structure is still not fully optimized."
        )
        print("This may lead to inaccurate results.")
        if optfreqissue:
            print(
                "In your case, this may be due to running an opt freq calculation using a single input file and IOP options."
            )
            print(
                "Gaussian does not carry the IOP options to the frequency calculation when using a single input file."
            )
            print(
                "To avoid this issue, run the optimization and frequency calculations separately."
            )
        else:
            print("To learn more about this issue, check https://gaussian.com/faq3/ .")
        print("Proceed at your own risk.")
        print("*" * 50)
        print("\n")


###############################################################


##FETCHES REFRACTIVE INDEX#####################################
def get_nr():
    buscar = False
    coms = [
        file
        for file in os.listdir("Geometries")
        if "Geometr" in file and ".com" in file
    ]
    with open("Geometries/" + coms[0], "r",encoding='utf-8') as f:
        for line in f:
            if "SCRF" in line.upper():
                buscar = True
                break
    if buscar:
        logs = [
            file
            for file in os.listdir("Geometries")
            if "Geometr" in file and ".log" in file
        ]
        for log in logs:
            with open("Geometries/" + log, "r",encoding='utf-8') as f:
                for line in f:
                    if "Solvent" in line and "Eps" in line:
                        line = line.split()
                        nr = np.sqrt(float(line[6]))
                        return nr
    else:
        return 1


###############################################################


##FETCHES CHARGE AND MULTIPLICITY##############################
def get_cm(freqlog):
    with open(freqlog, "r",encoding='utf-8') as f:
        for line in f:
            if "Charge" in line and "Multiplicity" in line:
                line = line.split()
                charge = line[2]
                mult = line[5]
                break
    return charge + " " + mult


###############################################################



#!/usr/bin/env python3
import os
import sys
import shutil
import numpy as np
import lx.tools
import lx.parser

##SOME CONSTANTS##############################################
EPSILON_0 = lx.parser.EPSILON_0  # F/m
HBAR_EV = lx.parser.HBAR_EV  # eV s
HBAR_J = lx.parser.HBAR_J  # J s
MASS_E = lx.parser.MASS_E  # kg
LIGHT_SPEED = lx.parser.LIGHT_SPEED  # m/s
E_CHARGE = lx.parser.E_CHARGE  # C
BOLTZ_EV = lx.parser.BOLTZ_EV  # eV/K
AMU = lx.parser.AMU  # kg
###############################################################


##SAMPLES GEOMETRIES###########################################
def make_geoms(freqlog, num_geoms, temp, header, bottom):
    lista = []
    counter = lx.tools.start_counter()
    _, atomos, structures = lx.tools.sample_geometries(
        freqlog, num_geoms, temp, 3000, warning=False
    )
    for n in range(np.shape(structures)[2]):
        final_geom = structures[:, :, n]
        lx.tools.write_input(
            atomos,
            final_geom,
            header.replace("UUUUU", str(n + 1)),
            bottom.replace("UUUUU", str(n + 1)),
            f"Geometry-{n+1+counter}-.com",
        )
        lista.append(f"Geometry-{n+1+counter}-.com")
    return lista


###############################################################


##GETS ENERGY FROM THE ORIGINAL FREQ LOG FILE##################
def get_energy_origin(freqlog):
    exc = 0
    with open(freqlog, "r",encoding='utf-8') as f:
        for line in f:
            if "SCF Done:" in line:
                line = line.split()
                scf = float(line[4]) * 27.2114
            elif "Total Energy," in line:
                line = line.split()
                exc = float(line[4]) * 27.2114
            elif "Rotational constants" in line:
                line = line.split()
                rot = [float(line[3]), float(line[4]), float(line[5])]
            elif "Normal termination" in line:
                if exc != 0:
                    scf = exc
                return scf, np.array([rot[0], rot[1], rot[2]])


###############################################################


##GETS ENERGIES FROM OPT LOG FILES#############################
def get_energies(folder, original_molecule):
    nums, scfs, rots = [], [], []
    files = [i for i in os.listdir(folder) if ".log" in i and "Geometry" in i]
    for file in files:
        exc = 0
        if np.array_equal(original_molecule, lx.tools.fingerprint(file, folder)):
            with open(folder + "/" + file, "r",encoding='utf-8') as f:
                num = float(file.split("-")[1])
                for line in f:
                    if "SCF Done:" in line:
                        line = line.split()
                        scf = float(line[4]) * 27.2114
                    elif "Total Energy," in line:
                        line = line.split()
                        exc = float(line[4]) * 27.2114
                    elif "Rotational constants" in line:
                        line = line.split()
                        rot = [float(line[3]), float(line[4]), float(line[5])]
                    elif "Normal termination" in line:
                        if exc != 0:
                            scf = exc
                        scfs.append(scf)
                        nums.append(num)
                        rots.append(rot)

    for file in files:
        try:
            shutil.move(file, "Geometries/" + file)
            shutil.move(file[:-3] + "com", "Geometries/" + file[:-3] + "com")
        except FileNotFoundError:
            pass
    return np.array(nums), np.array(scfs), np.array(rots)


###############################################################


def measure(vec1, vec2, cr):
    vec1 = np.array(vec1)
    vec2 = np.array(vec2)
    cr = np.array(cr)
    dist = max(abs(vec1 - vec2) - cr)
    distance = np.heaviside(dist, 0)
    return distance


class Conformation:
    def __init__(self, rot, energy, identity, num) -> None:
        self.rot = rot[np.newaxis, :]
        self.energy = [energy]
        self.identity = identity
        self.std = rot / 1000
        self.num = [num]

    def add_rot(self, rot, num, energy):
        self.rot = np.vstack((self.rot, rot[np.newaxis, :]))
        newstd = np.std(self.rot, axis=0)
        # get higher std
        self.std = np.where(newstd > self.std, newstd, self.std)
        self.num.append(num)
        self.energy.append(energy)

    def merge_rot(self, rot):
        self.rot = np.vstack((self.rot, rot))
        newstd = np.std(self.rot, axis=0)
        # get higher std
        self.std = np.where(newstd > self.std, newstd, self.std)

    def get_avg(self):
        return np.mean(self.rot, axis=0)


def internal_comparison(conformations):
    num_conformations = len(conformations)
    remove = set()  # Use a set to avoid duplicates

    for i in range(num_conformations - 1):
        if i in remove:
            continue  # Skip already marked for removal
        for j in range(i + 1, num_conformations):
            if j in remove:
                continue  # Skip already marked for removal
            distance = measure(
                conformations[i].get_avg(),
                conformations[j].get_avg(),
                conformations[j].std + conformations[i].std,
            )
            if distance == 0:
                conformations[i].num += conformations[j].num
                conformations[i].energy += conformations[j].energy
                conformations[i].merge_rot(conformations[j].rot)
                remove.add(j)
                
    # Remove elements from list whose index is in remove
    conformations = [c for idx, c in enumerate(conformations) if idx not in remove]
    return conformations


def classify(conformations, folder):
    nums, scfs, rots = get_energies(folder, conformations[0].identity)
    new_conformations = []

    total_geometries = rots.shape[0]
    matched_count = 0
    new_conformation_count = 0

    for i in range(total_geometries):
        matched = False
        for conformation in conformations:
            distance = measure(rots[i, :], conformation.get_avg(), conformation.std)
            if distance == 0:
                conformation.add_rot(rots[i, :], nums[i], scfs[i])
                matched = True
                matched_count += 1
                break

        if not matched:
            new_conformations.append(Conformation(rots[i, :], scfs[i], conformations[0].identity, nums[i]))
            new_conformation_count += 1

    conformations.extend(new_conformations)

    if len(conformations) > 1:
        conformations = internal_comparison(conformations)
    return conformations



def write_report(conformations, rounding, total_rounds, temp):
    engs = []
    for conformation in conformations:
        engs.append(np.mean(conformation.energy))
    engs = np.array(engs)
    argsort = np.argsort(engs)
    engs = engs[argsort]
    # sort conformations as list
    conformations = [conformations[i] for i in argsort]
    deltae = engs - min(engs)
    probs = np.exp(-(deltae) / (0.026))
    probs = probs / sum(probs)

    with open("conformation.lx", "w",encoding='utf-8') as f:
        f.write(
            f"{'#Group':<6}  {'Energy(eV)':<10}  {'DeltaE(eV)':<10}  {'Prob@300K(%)':<12}  {'Rot1':<10}  {'Rot2':<10}  {'Rot3':<10}  {'Std1':<10}  {'Std2':<10}  {'Std3':<10}  {'Number':<6}  {'Last':<6}\n"
        )
        for i, conformation in enumerate(conformations):
            rot = conformation.get_avg()
            std = conformation.std
            last = conformation.num[-1]
            total = len(conformation.num)
            f.write(
                f"{i+1:<6}  {engs[i]:<10.3f}  {deltae[i]:<10.3f}  {100*probs[i]:<12.1f}  {rot[0]:<10.7f}  {rot[1]:<10.7f}  {rot[2]:<10.7f}  {std[0]:<10.7f}  {std[1]:<10.7f}  {std[2]:<10.7f}  {total:<6.0f}  {last:<6.0f}\n"
            )
        f.write(f"\n#Round {rounding}/{total_rounds} Temperature: {temp} K")


##RUNS FREQ CALCULATION FOR NEW CONFORMATION###################
def rodar_freq(origin, nproc, mem, base, cm, batch_file, gaussian):
    geomlog = f"Geometries/Geometry-{origin:.0f}-.log"
    geom, atomos = lx.parser.pega_geom(geomlog)
    header = f"%nproc={nproc}\n%mem={mem}\n# freq=(noraman) nosymm  {base} \n\nTITLE\n\n{cm}\n"
    file = f"Freq-{origin:.0f}-.com"
    lx.tools.write_input(atomos, geom, header, "", file)
    the_watcher = lx.tools.Watcher('.',files=[file])
    the_watcher.run(batch_file, gaussian, 1)
    the_watcher.hold_watch()
    #lx.tools.rodar_lista([file], batch_file, gaussian, "conformation.lx", 1)
    log = file[:-3] + "log"
    with open(log, "r",encoding='utf-8') as f:
        for line in f:
            if "Normal termination" in line:
                return log
            elif "Error termination" in line:
                return None


###############################################################


def classify_only():
    files = [i for i in os.listdir(".") if "Geometry-" in i and ".log" in i]
    conformations = []
    for file in files:
        num = int(file.split("-")[1])
        scf, rot = get_energy_origin(file)
        conformations.append(
            Conformation(rot, scf, lx.tools.fingerprint(file, "."), num)
        )
    conformations = internal_comparison(conformations)
    write_report(conformations, 0, 0, 0)


def main():
    freqlog = sys.argv[1]
    base = sys.argv[2]
    nproc = sys.argv[3]
    mem = sys.argv[4]
    temperature = float(sys.argv[5])
    delta_t = float(sys.argv[6])
    num_geoms = int(sys.argv[7])
    rounds = int(sys.argv[8])
    numjobs = int(sys.argv[9])
    script = sys.argv[10]
    gaussian = sys.argv[11]
    temp_0 = temperature
    freq0 = freqlog
    if "td" in base.lower():
        opt = "=loose"
    else:
        opt = ""

    try:
        os.mkdir("Geometries")
    except FileExistsError:
        pass
    cm = lx.parser.get_cm(freqlog)
    header = (
        f"%nproc={nproc}\n%mem={mem}\n# opt{opt} nosymm  {base} \n\nTitle\n\n{cm}\n"
    )
    scf, rot = get_energy_origin(freqlog)
    conformations = [Conformation(rot, scf, lx.tools.fingerprint(freqlog, "."), 0)]
    files = [i for i in os.listdir("Geometries") if "Geometry" in i and ".log" in i]
    if len(files) > 0:
        conformations = classify(conformations, "Geometries")
        write_report(conformations, 0, rounds, temp_0)
    else:
        pass

    groups = len(conformations)
    for i in range(rounds):
        lista = make_geoms(freqlog, num_geoms, temp_0, header, "")
        the_watcher = lx.tools.Watcher('.',files=lista)
        the_watcher.run(script, gaussian, numjobs)
        the_watcher.hold_watch()
        #lx.tools.rodar_lista(lista, script, gaussian, "conformation.lx", numjobs)
        conformations = classify(conformations, ".")
        write_report(conformations, i + 1, rounds, temp_0)

        if len(conformations) != groups:
            log = rodar_freq(
                conformations[-1].num[-1], nproc, mem, base, cm, script, gaussian
            )
            if log is not None:
                freqlog = log
                temp_0 = temperature
            groups = len(conformations)
        else:
            temp_0 += delta_t

    with open("conformation.lx", "a",encoding='utf-8') as f:
        f.write("\n#Search concluded!")

    try:
        os.mkdir("Conformers")
    except FileExistsError:
        pass

    for i, conformation in enumerate(conformations):
        numero = conformation.num[-1]
        if numero == 0:
            freqlog = freq0
        else:
            freqlog = f"Geometries/Geometry-{numero:.0f}-.log"
        _, _, nproc, mem, scrf, _ = lx.parser.busca_input(freqlog)
        cm = lx.parser.get_cm(freqlog)
        header = f"%nproc={nproc}\n%mem={mem}\n%chk=Group_{i+1}_.chk\n# pm6 {scrf} opt nosymm\n\nTITLE\n\n{cm}\n"
        geom, atomos = lx.parser.pega_geom(freqlog)
        lx.tools.write_input(
            atomos, geom, header, "", f"Conformers/Geometry-{i+1}-.com"
        )


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
import os
import sys
import shutil
import lx.tools

BATCH_FILE = sys.argv[1]
GAUSSIAN = sys.argv[2]
NUM = int(sys.argv[3])

def run_batch():
    scripts = [i for i in os.listdir(".") if ".sh" in i]
    for file in scripts:
        shutil.copy(file, "Geometries")
    os.chdir("Geometries")
    files = [i for i in os.listdir(".") if "Geometr" in i and ".com" in i]
    factor = lx.tools.set_factor(files[0])
    the_watcher = lx.tools.Watcher('.', counter=factor)
    the_watcher.run(BATCH_FILE, GAUSSIAN, NUM)
